=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2010-01-11 02:07:30 +0000
+++ bzrlib/errors.py	2010-01-11 05:58:22 +0000
@@ -704,11 +704,12 @@
 
     _fmt = 'Not a branch: "%(path)s"%(detail)s.'
 
-    def __init__(self, path, bzrdir=None):
+    def __init__(self, path, detail=None, bzrdir=None):
        import bzrlib.urlutils as urlutils
        path = urlutils.unescape_for_display(path, 'ascii')
-       # remember the detail in case of remote serialization
-       self.detail = None
+       if detail is not None:
+           detail = ': ' + detail
+       self.detail = detail
        self.bzrdir = bzrdir
        PathError.__init__(self, path=path)
 

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2010-01-08 06:08:54 +0000
+++ bzrlib/remote.py	2010-01-11 05:58:22 +0000
@@ -284,21 +284,32 @@
     def _get_branch_reference(self):
         path = self._path_for_remote_call(self._client)
         medium = self._client._medium
-        if not medium._is_remote_before((1, 13)):
+        candidate_calls = [
+            ('BzrDir.open_branchV3', (2, 1)),
+            ('BzrDir.open_branchV2', (1, 13)),
+            ('BzrDir.open_branch', None),
+            ]
+        for verb, required_version in candidate_calls:
+            if required_version and medium._is_remote_before(required_version):
+                continue
             try:
-                response = self._call('BzrDir.open_branchV2', path)
-                if response[0] not in ('ref', 'branch'):
-                    raise errors.UnexpectedSmartServerResponse(response)
-                return response
+                response = self._call(verb, path)
             except errors.UnknownSmartMethod:
-                medium._remember_remote_is_before((1, 13))
-        response = self._call('BzrDir.open_branch', path)
-        if response[0] != 'ok':
+                if required_version is None:
+                    raise
+                medium._remember_remote_is_before(required_version)
+            else:
+                break
+        if verb == 'BzrDir.open_branch':
+            if response[0] != 'ok':
+                raise errors.UnexpectedSmartServerResponse(response)
+            if response[1] != '':
+                return ('ref', response[1])
+            else:
+                return ('branch', '')
+        if response[0] not in ('ref', 'branch'):
             raise errors.UnexpectedSmartServerResponse(response)
-        if response[1] != '':
-            return ('ref', response[1])
-        else:
-            return ('branch', '')
+        return response
 
     def _get_tree_branch(self):
         """See BzrDir._get_tree_branch()."""
@@ -2823,11 +2834,13 @@
         raise NoSuchRevision(find('branch'), err.error_args[0])
     elif err.error_verb == 'nosuchrevision':
         raise NoSuchRevision(find('repository'), err.error_args[0])
-    elif err.error_tuple == ('nobranch',):
-        raise errors.NotBranchError(path=find('bzrdir').root_transport.base)
     elif err.error_verb == 'nobranch':
+        if len(err.error_args) >= 1:
+            extra = err.error_args[0]
+        else:
+            extra = None
         raise errors.NotBranchError(path=find('bzrdir').root_transport.base,
-            detail=err.error_args[0])
+            detail=extra)
     elif err.error_verb == 'norepository':
         raise errors.NoRepositoryPresent(find('bzrdir'))
     elif err.error_verb == 'LockContention':

=== modified file 'bzrlib/smart/bzrdir.py'
--- bzrlib/smart/bzrdir.py	2009-10-14 20:14:15 +0000
+++ bzrlib/smart/bzrdir.py	2010-01-11 05:58:22 +0000
@@ -89,9 +89,7 @@
             self._bzrdir = BzrDir.open_from_transport(
                 self.transport_from_client_path(path))
         except errors.NotBranchError, e:
-            if e.detail is None:
-                return FailedSmartServerResponse(('nobranch',))
-            return FailedSmartServerResponse(('nobranch', e.detail))
+            return FailedSmartServerResponse(('nobranch',))
         return self.do_bzrdir_request(*args)
 
     def _boolean_to_yes_no(self, a_boolean):
@@ -468,9 +466,7 @@
             else:
                 return SuccessfulSmartServerResponse(('ok', reference_url))
         except errors.NotBranchError, e:
-            if e.detail is None:
-                return FailedSmartServerResponse(('nobranch',))
-            return FailedSmartServerResponse(('nobranch', e.detail))
+            return FailedSmartServerResponse(('nobranch',))
 
 
 class SmartServerRequestOpenBranchV2(SmartServerRequestBzrDir):
@@ -486,6 +482,36 @@
             else:
                 return SuccessfulSmartServerResponse(('ref', reference_url))
         except errors.NotBranchError, e:
-            if e.detail is None:
-                return FailedSmartServerResponse(('nobranch',))
-            return FailedSmartServerResponse(('nobranch', e.detail))
+            return FailedSmartServerResponse(('nobranch',))
+
+
+class SmartServerRequestOpenBranchV3(SmartServerRequestBzrDir):
+
+    def do_bzrdir_request(self):
+        """Open a branch at path and return the reference or format.
+        
+        This version introduced in 2.1.
+
+        Differences to SmartServerRequestOpenBranchV2:
+          * can return 2-element ('nobranch', extra), where 'extra' is a string
+            with an explanation like 'location is a repository'.  Previously
+            a 'nobranch' response would never have more than one element.
+        """
+        try:
+            reference_url = self._bzrdir.get_branch_reference()
+            if reference_url is None:
+                br = self._bzrdir.open_branch(ignore_fallbacks=True)
+                format = br._format.network_name()
+                return SuccessfulSmartServerResponse(('branch', format))
+            else:
+                return SuccessfulSmartServerResponse(('ref', reference_url))
+        except errors.NotBranchError, e:
+            str(e)
+            resp = ('nobranch',)
+            detail = e.detail
+            if detail:
+                if detail.startswith(': '):
+                    detail = detail[2:]
+                resp += (detail,)
+            return FailedSmartServerResponse(resp)
+

=== modified file 'bzrlib/smart/request.py'
--- bzrlib/smart/request.py	2009-12-21 17:00:29 +0000
+++ bzrlib/smart/request.py	2010-01-11 05:58:22 +0000
@@ -561,6 +561,9 @@
     'BzrDir.open_branchV2', 'bzrlib.smart.bzrdir',
     'SmartServerRequestOpenBranchV2')
 request_handlers.register_lazy(
+    'BzrDir.open_branchV3', 'bzrlib.smart.bzrdir',
+    'SmartServerRequestOpenBranchV3')
+request_handlers.register_lazy(
     'delete', 'bzrlib.smart.vfs', 'DeleteRequest')
 request_handlers.register_lazy(
     'get', 'bzrlib.smart.vfs', 'GetRequest')

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2009-12-16 22:29:31 +0000
+++ bzrlib/tests/test_remote.py	2010-01-11 05:58:22 +0000
@@ -440,7 +440,7 @@
             'BzrDir.cloning_metadir', ('quack/', 'False'),
             'error', ('BranchReference',)),
         client.add_expected_call(
-            'BzrDir.open_branchV2', ('quack/',),
+            'BzrDir.open_branchV3', ('quack/',),
             'success', ('ref', self.get_url('referenced'))),
         a_bzrdir = RemoteBzrDir(transport, remote.RemoteBzrDirFormat(),
             _client=client)
@@ -531,7 +531,7 @@
         self.make_branch('.')
         a_dir = BzrDir.open(self.get_url('.'))
         self.reset_smart_call_log()
-        verb = 'BzrDir.open_branchV2'
+        verb = 'BzrDir.open_branchV3'
         self.disable_verb(verb)
         format = a_dir.open_branch()
         call_count = len([call for call in self.hpss_calls if
@@ -547,7 +547,7 @@
         transport = transport.clone('quack')
         client = FakeClient(transport.base)
         client.add_expected_call(
-            'BzrDir.open_branchV2', ('quack/',),
+            'BzrDir.open_branchV3', ('quack/',),
             'success', ('branch', branch_network_name))
         client.add_expected_call(
             'BzrDir.find_repositoryV3', ('quack/',),
@@ -572,7 +572,7 @@
             _client=client)
         self.assertRaises(errors.NotBranchError, bzrdir.open_branch)
         self.assertEqual(
-            [('call', 'BzrDir.open_branchV2', ('quack/',))],
+            [('call', 'BzrDir.open_branchV3', ('quack/',))],
             client._calls)
 
     def test__get_tree_branch(self):
@@ -602,7 +602,7 @@
         network_name = reference_format.network_name()
         branch_network_name = self.get_branch_format().network_name()
         client.add_expected_call(
-            'BzrDir.open_branchV2', ('~hello/',),
+            'BzrDir.open_branchV3', ('~hello/',),
             'success', ('branch', branch_network_name))
         client.add_expected_call(
             'BzrDir.find_repositoryV3', ('~hello/',),
@@ -1190,7 +1190,7 @@
         client = FakeClient(self.get_url())
         branch_network_name = self.get_branch_format().network_name()
         client.add_expected_call(
-            'BzrDir.open_branchV2', ('stacked/',),
+            'BzrDir.open_branchV3', ('stacked/',),
             'success', ('branch', branch_network_name))
         client.add_expected_call(
             'BzrDir.find_repositoryV3', ('stacked/',),
@@ -1226,7 +1226,7 @@
         client = FakeClient(self.get_url())
         branch_network_name = self.get_branch_format().network_name()
         client.add_expected_call(
-            'BzrDir.open_branchV2', ('stacked/',),
+            'BzrDir.open_branchV3', ('stacked/',),
             'success', ('branch', branch_network_name))
         client.add_expected_call(
             'BzrDir.find_repositoryV3', ('stacked/',),
@@ -2775,6 +2775,15 @@
         expected_error = errors.NotBranchError(path=bzrdir.root_transport.base)
         self.assertEqual(expected_error, translated_error)
 
+    def test_nobranch_one_arg(self):
+        bzrdir = self.make_bzrdir('')
+        translated_error = self.translateTuple(
+            ('nobranch', 'extra detail'), bzrdir=bzrdir)
+        expected_error = errors.NotBranchError(
+            path=bzrdir.root_transport.base,
+            detail='extra detail')
+        self.assertEqual(expected_error, translated_error)
+
     def test_LockContention(self):
         translated_error = self.translateTuple(('LockContention',))
         expected_error = errors.LockContention('(remote lock)')

=== modified file 'bzrlib/tests/test_smart.py'
--- bzrlib/tests/test_smart.py	2010-01-08 06:08:54 +0000
+++ bzrlib/tests/test_smart.py	2010-01-11 05:58:22 +0000
@@ -529,7 +529,7 @@
         backing = self.get_transport()
         request = smart.bzrdir.SmartServerRequestOpenBranch(backing)
         repo = self.make_repository('.')
-        self.assertEqual(SmartServerResponse(('nobranch', 'location is a repository')),
+        self.assertEqual(SmartServerResponse(('nobranch',)),
             request.execute(''))
 
 
@@ -588,7 +588,67 @@
         backing = self.get_transport()
         request = smart.bzrdir.SmartServerRequestOpenBranchV2(backing)
         repo = self.make_repository('.')
-        self.assertEqual(SmartServerResponse(('nobranch', 'location is a repository')),
+        self.assertEqual(SmartServerResponse(('nobranch',)),
+            request.execute(''))
+
+
+class TestSmartServerRequestOpenBranchV3(TestCaseWithChrootedTransport):
+
+    def test_no_branch(self):
+        """When there is no branch, ('nobranch', ) is returned."""
+        backing = self.get_transport()
+        self.make_bzrdir('.')
+        request = smart.bzrdir.SmartServerRequestOpenBranchV3(backing)
+        self.assertEqual(SmartServerResponse(('nobranch',)),
+            request.execute(''))
+
+    def test_branch(self):
+        """When there is a branch, 'ok' is returned."""
+        backing = self.get_transport()
+        expected = self.make_branch('.')._format.network_name()
+        request = smart.bzrdir.SmartServerRequestOpenBranchV3(backing)
+        self.assertEqual(SuccessfulSmartServerResponse(('branch', expected)),
+            request.execute(''))
+
+    def test_branch_reference(self):
+        """When there is a branch reference, the reference URL is returned."""
+        self.vfs_transport_factory = local.LocalURLServer
+        backing = self.get_transport()
+        request = smart.bzrdir.SmartServerRequestOpenBranchV3(backing)
+        branch = self.make_branch('branch')
+        checkout = branch.create_checkout('reference',lightweight=True)
+        reference_url = BranchReferenceFormat().get_reference(checkout.bzrdir)
+        self.assertFileEqual(reference_url, 'reference/.bzr/branch/location')
+        self.assertEqual(SuccessfulSmartServerResponse(('ref', reference_url)),
+            request.execute('reference'))
+
+    def test_stacked_branch(self):
+        """Opening a stacked branch does not open the stacked-on branch."""
+        trunk = self.make_branch('trunk')
+        feature = self.make_branch('feature')
+        feature.set_stacked_on_url(trunk.base)
+        opened_branches = []
+        Branch.hooks.install_named_hook('open', opened_branches.append, None)
+        backing = self.get_transport()
+        request = smart.bzrdir.SmartServerRequestOpenBranchV3(backing)
+        request.setup_jail()
+        try:
+            response = request.execute('feature')
+        finally:
+            request.teardown_jail()
+        expected_format = feature._format.network_name()
+        self.assertEqual(
+            SuccessfulSmartServerResponse(('branch', expected_format)),
+            response)
+        self.assertLength(1, opened_branches)
+
+    def test_notification_on_branch_from_repository(self):
+        """When there is a repository, the error should return details."""
+        backing = self.get_transport()
+        request = smart.bzrdir.SmartServerRequestOpenBranchV3(backing)
+        repo = self.make_repository('.')
+        self.assertEqual(
+            SmartServerResponse(('nobranch', 'location is a repository')),
             request.execute(''))
 
 
@@ -1787,6 +1847,8 @@
             smart.bzrdir.SmartServerRequestOpenBranch)
         self.assertHandlerEqual('BzrDir.open_branchV2',
             smart.bzrdir.SmartServerRequestOpenBranchV2)
+        self.assertHandlerEqual('BzrDir.open_branchV3',
+            smart.bzrdir.SmartServerRequestOpenBranchV3)
         self.assertHandlerEqual('PackRepository.autopack',
             smart.packrepository.SmartServerPackRepositoryAutopack)
         self.assertHandlerEqual('Repository.gather_stats',

